"""Core package

This sub-package contains the principal modules, which are responsible for the
creation of COBRApy objects and the addition of Pathway objects into the
metabolic model. The modules include:

- creation: Creation of COBRApy objects
- crossreferences: Storing and retriving crossreferences information to COBRApy
  objects
- extension: Addition of reactions and metabolites into metabolic models
- genes: Creation of COBRApy genes
- graph: Algorithm to build the order of reactions
- pathway: Creation of :func:`cobra.core.group.Group` child.
:func:`cobramod.core.pathway.Pathway`.
- summary: Short and long information about changes
"""
