"""Sub-package parsing

This sub-package takes care of gathering, storing and parsing the data from
multiple servers

- bigg: Parsing of BiGG information
- biocyc: Parsing and retrieval of BioCyc families
- db_version: Database versioning configurator
- kegg: Parsing of KEGG information
- plantcyc: Gene parsing for Plantcyc
- solcyc: Gene parsing for SolCyc
"""
