"""Sub-package visualization

This package handles the creation and parsing of a JSON that can be used for
the package Escher. The main module converter.py has the class JsonDictionary,
which is the base for the Escher-visualization. The module items.py, is a
pythonic-representation of the JSON-objects from the Escher-schema, which
includes:

- Reactions
- Segments
- Nodes

Moreover, there is a debug module, which helps debugging the objects and the
initialization of the module. Please refer to each individual module for more
in-depth information.
"""
